"""Cobertura aggregator tests"""
import os
import unittest
import urllib2
from mock import patch, MagicMock
from base import CoberturaJSONAggregator

API_SETTINGS = {
    CoberturaJSONAggregator.USERNAME_KEY: "test",
    CoberturaJSONAggregator.API_TOKEN_KEY: "test",
    CoberturaJSONAggregator.DOMAIN_KEY: "http://localhost",
    CoberturaJSONAggregator.JOBS_KEY: ["test"],
    CoberturaJSONAggregator.TARGETS_KEY: [
        'club',
        'cron',
    ],
}


class CoberturaJSONAggregatorTests(unittest.TestCase):
    """Test cases for aggregating Cobertura JSON via Jenkins REST API"""
    def setUp(self):
        self.test_name = 'test'
        self.aggregator = CoberturaJSONAggregator(self.test_name, API_SETTINGS)
        fixture_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'fixtures/cobertura.json'
        )
        with open(fixture_path) as fixture_file:
            self._cobertura_json = fixture_file.read()

    @patch('base.urllib2.urlopen')
    def test_report_generator(self, urlopen_mock):
        urlopen_mock().read.return_value = self._cobertura_json
        self.aggregator.generate_report()
        expected_result = [
            ['club', '67.50%', '62.50%'],
            ['cron', '8.15%', '6.25%']
        ]
        self.assertEqual(self.aggregator._report, expected_result)

    @patch('base.urllib2.urlopen')
    @patch('base.LOGGER')
    def test_http_error(self, logger_mock, urlopen_mock):
        """Test urllib2.HTTPError exceptions"""
        log = []
        logger_mock.error.side_effect = log.append
        logger_mock.info.side_effect = log.append

        urlopen_mock.side_effect = urllib2.HTTPError(
            "test",
            404,
            "test",
            {},
            MagicMock()
        )
        self.aggregator.generate_report()
        http_error_results = [
            'http://localhost/job/test/lastSuccessfulBuild/cobertura/api/json?depth=4\nHTTP error: HTTP Error 404: test'
        ]
        self.assertEqual(log, http_error_results)

    @patch('base.urllib2.urlopen')
    @patch('base.LOGGER')
    def test_url_error(self, logger_mock, urlopen_mock):
        """Test urllib2.URLError exceptions"""
        log = []
        logger_mock.error.side_effect = log.append
        logger_mock.info.side_effect = log.append

        urlerror_results = [
            'http://localhost/job/test/lastSuccessfulBuild/cobertura/api/json?depth=4\nURL error: URL Error'
        ]
        urlopen_mock.side_effect = urllib2.URLError("URL Error")
        self.aggregator.generate_report()
        self.assertEqual(log, urlerror_results)

    @patch('base.urllib2.urlopen')
    @patch('base.LOGGER')
    def test_simplejson_error(self, logger_mock, urlopen_mock):
        """Test simplejson.loads error"""
        log = []
        logger_mock.error.side_effect = log.append
        logger_mock.info.side_effect = log.append

        urlopen_mock().read.return_value = '{'
        self.aggregator.generate_report()
        simplejson_error = [
            "{\nCobertura URL did not return valid json: Expecting property name enclosed in double quotes or '}': line 1 column 2 (char 1)"
        ]
        self.assertEqual(log, simplejson_error)



"""Cobertura aggregator tests"""
import os
import unittest
from base import CoberturaXMLAggregator

XML_SETTINGS = {
    CoberturaXMLAggregator.XML_FILES_KEY: [os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'fixtures/cobertura.xml'
        )],
    CoberturaXMLAggregator.TARGETS_KEY: [
        'mytests',
        'widget_core/v3/core-model.js',
        'widget_core/common/getElementsByRegex.js'
    ]
}


class CoberturaXMLAggregatorTests(unittest.TestCase):
    """Test cases for agggreating Cobertura XML"""
    def setUp(self):
        self.test_name = 'tests'
        self.aggregator = CoberturaXMLAggregator(self.test_name, XML_SETTINGS)

    def test_report_generator(self):
        """Test generating a report"""
        self.aggregator.generate_report()
        expected_result = [
            ['mytests', '73.33%', '48.65%'],
            ['widget_core/v3/core-model.js', '84.44%', '77.01%'],
            ['widget_core/common/getElementsByRegex.js', '100.00%', '100.00%']
        ]
        self.assertEqual(self.aggregator._report, expected_result)

    def test_print_report(self):
        """Test report generation"""
        settings = XML_SETTINGS
        settings[self.aggregator.REPORT_PATH_KEY] = '/tmp'
        self.aggregator.set_settings(self.test_name, XML_SETTINGS)

        self.aggregator.generate_report()
        self.aggregator.print_report()

        report_file = self.aggregator.get_report_file()
        with open(report_file) as report:
            self.assertEqual(report.read(),
                             """Target                                    Line%    Branch%
----------------------------------------  -------  ---------
mytests                                   73.33%   48.65%
widget_core/v3/core-model.js              84.44%   77.01%
widget_core/common/getElementsByRegex.js  100.00%  100.00%"""
                             )
        os.unlink(report_file)



import unittest
from run import run
from mock import patch, Mock


class CLIRunnerTests(unittest.TestCase):
    """Test suite for aggregator CLI runner"""
    @patch('run.imp')
    @patch('run.CoberturaJSONAggregator')
    @patch('run.CoberturaXMLAggregator')
    @patch('run.argparse')
    def test_runner(self, argparse_mock, xml_mock, api_mock, imp_mock):
        """XML and API type running configs"""
        mock_settings = Mock()
        mock_settings.SETTINGS = {'test': {'TYPE': 'xml'}}
        imp_mock.load_source.return_value = mock_settings
        xml_mock.TYPE_KEY = 'TYPE'
        run('')

        xml_mock().set_settings.assert_called_once_with(
            'test', {'TYPE': 'xml'}
        )
        xml_mock().generate_report.assert_called_once()
        xml_mock().print_report.assert_called_once()

        mock_settings.SETTINGS = {'test': {'TYPE': 'jenkins_api'}}
        imp_mock.load_source.return_value = mock_settings
        run('')

        api_mock().set_settings.assert_called_once_with(
            'test', {'TYPE': 'jenkins_api'}
        )
        api_mock().generate_report.assert_called_once()
        api_mock().print_report.assert_called_once()

        mock_settings.SETTINGS = {'test': {'TYPE': 'test'}}
        imp_mock.load_source.return_value = mock_settings
        self.assertRaises(UnboundLocalError, run, '')



"""Cobertura report aggregator for a given set of TARGETS"""
import base64
import calendar
import logging
import simplejson
import time
import urllib2
import xml.etree.ElementTree as ET
from tabulate import tabulate

LOGGER = logging.getLogger()


class CoberturaAggregator(object):
    """Cobertura aggregation base class"""
    REPORT_PATH_KEY = 'REPORT_PATH'
    TARGETS_KEY = 'TARGETS'
    TYPE_KEY = 'TYPE'

    def __init__(self, name=None, settings={}):
        self._name = name
        self._report_path = settings.get(self.REPORT_PATH_KEY, '')
        self._targets = settings.get(self.TARGETS_KEY, [])
        self._report = []
        self._target_stats = {}
        for target in self._targets:
            self._target_stats[target] = {
                'lines': 0,
                'lines_missed': 0,
                'cond': 0,
                'cond_missed': 0,
                'line_coverage': float(0),
                'cond_coverage': float(0),
            }

        if self._report_path and self._report_path[-1] != '/':
            self._report_path = '{}/'.format(self._report_path)

        # /path/to/file/1387487334_cobertura_agg.txt
        self._report_file = '{}{}_cobertura_agg.txt'.format(
            self._report_path,
            calendar.timegm(time.gmtime())
            )

    def set_settings(self, name, settings):
        """Reset instance settings"""
        self.__init__(name, settings)

    def get_report_file(self):
        """Get report file path"""
        return self._report_file

    def print_report(self):
        """Print the report"""
        headers = ['Target', 'Line%', 'Branch%']
        table = tabulate(self._report, headers)
        print """

---------
{} REPORT
---------
""".format(self._name)
        print table

        if self._report_path:
            with open(self._report_file, 'w+') as report_file:
                report_file.write(table)

    def generate_report(self):
        """Each aggregator needs to implement this"""
        raise NotImplementedError

    def _get_stats(self, **kwargs):
        """Search for target aggregations"""
        raise NotImplementedError

    def _get_target_stats(self, target, classes):
        """Update target aggregation"""
        raise NotImplementedError

    def _dump_stats(self):
        """Dump aggregations into a useable list for tabulate"""
        for target in self._targets:
            stats = self._target_stats[target]
            self._report.append([
                target,
                "{:.2f}%".format(stats['line_coverage']),
                "{:.2f}%".format(stats['cond_coverage'])
                ]
            )

    def _update_target_stats(self,
                             target,
                             lines_covered,
                             total_lines,
                             branches_covered,
                             total_branches
                             ):
        """Update target stats"""
        stats = self._target_stats[target]
        stats['lines'] += total_lines
        stats['lines_missed'] += (total_lines - lines_covered)
        if total_lines:
            perc = float(stats['lines_missed']) / float(stats['lines'])
            stats['line_coverage'] = 100 - (perc * 100)
        else:
            stats['line_coverage'] = stats['line_coverage'] or float(100)

        stats['cond'] += total_branches
        stats['cond_missed'] += (total_branches - branches_covered)
        if total_branches:
            perc = float(stats['cond_missed']) / float(stats['cond'])
            stats['cond_coverage'] = 100 - (perc * 100)
        else:
            stats['cond_coverage'] = stats['line_coverage'] or float(100)
        self._target_stats[target] = stats


class CoberturaJSONAggregator(CoberturaAggregator):
    """Cobertura REST JSON aggregator"""
    USERNAME_KEY = 'USERNAME'
    API_TOKEN_KEY = 'API_TOKEN'
    DOMAIN_KEY = 'DOMAIN'
    JOBS_KEY = 'JOBS'

    _LINES_KEY = 'Lines'
    _CONDITIONALS_KEY = 'Conditionals'
    _NUMERATOR_KEY = 'numerator'
    _DENOMINATOR_KEY = 'denominator'

    def __init__(self, name=None, settings={}):
        super(CoberturaJSONAggregator, self).__init__(name, settings)
        self._username = settings.get(self.USERNAME_KEY)
        self._api_token = settings.get(self.API_TOKEN_KEY)
        self._domain = settings.get(self.DOMAIN_KEY)
        self._jobs = settings.get(self.JOBS_KEY, [])

        coverage_uri = "lastSuccessfulBuild/cobertura/api/json?depth=4"
        self._cobertura_urls = []
        for job in self._jobs:
            self._cobertura_urls.append(
                "{}/job/{}/{}".format(self._domain, job, coverage_uri)
                )

    def generate_report(self):
        for cobertura_url in self._cobertura_urls:
            auth_header = 'Basic: ' + base64.b64encode('%s:%s' % (
                self._username,
                self._api_token
                )
            ).strip()
            headers = {'Authorization': auth_header}
            request = urllib2.Request(
                cobertura_url,
                headers=headers
            )

            try:
                resp = urllib2.urlopen(request, timeout=5)
                cobertura_json = resp.read()
                cobertura_report = simplejson.loads(cobertura_json)
            except urllib2.HTTPError as e:
                LOGGER.error("{}\nHTTP error: {}".format(cobertura_url, e))
                continue
            except urllib2.URLError as e:
                LOGGER.error("{}\nURL error: {}".format(
                    cobertura_url,
                    e.reason
                    )
                )
                continue
            except simplejson.scanner.JSONDecodeError as e:
                LOGGER.error(
                    "{}\nCobertura URL did not return valid json: {}".format(
                        resp.read(),
                        e
                    )
                )
                continue

            # Drill down to what we care about...
            results = cobertura_report.get('results')
            children = results.get('children')
            for child in children:
                self._get_stats(child.get('children'))

        self._dump_stats()

    def _get_stats(self, children):
        for child in children:
            for target in self._targets:
                if target in child.get('name'):
                    self._get_target_stats(target, child)

    def _get_target_stats(self, target, child):
        lines_covered = 0
        total_lines = 0
        branches_covered = 0
        total_branches = 0

        for element in child.get('elements'):
            name = element.get('name')
            if name == self._LINES_KEY:
                total_lines += element.get(self._DENOMINATOR_KEY)
                lines_covered += element.get(self._NUMERATOR_KEY)
            elif name == self._CONDITIONALS_KEY:
                total_branches += element.get(self._DENOMINATOR_KEY)
                branches_covered += element.get(self._NUMERATOR_KEY)

        self._update_target_stats(**{
            'target': target,
            'lines_covered': lines_covered,
            'total_lines': total_lines,
            'branches_covered': branches_covered,
            'total_branches': total_branches

        })


class CoberturaXMLAggregator(CoberturaAggregator):
    """Cobertura XML aggregator"""
    XML_FILES_KEY = 'XML_FILES'

    def __init__(self, name=None, settings={}):
        super(CoberturaXMLAggregator, self).__init__(name, settings)
        self._cobertura_xml = settings.get(self.XML_FILES_KEY, [])
        self._xml_tree = None

    def generate_report(self):
        for cobertura_report in self._cobertura_xml:
            xml_tree = ET.parse(cobertura_report)

            root = xml_tree.getroot()
            for classes in root.iter('classes'):
                self._get_stats(classes)

        self._dump_stats()

    def _get_stats(self, classes):
        """Search for target aggregations"""
        cobertura_class = classes.find('class')
        class_fn = cobertura_class.get('filename')
        for target in self._targets:
            if class_fn.find(target) == 0:
                self._get_target_stats(target, classes)

    def _get_target_stats(self, target, classes):
        """Update target aggregation"""
        lines_covered = 0
        total_lines = 0
        branches_covered = 0
        total_branches = 0

        for line in classes.iter('line'):
            total_lines += 1
            if int(line.get('hits')) > 0:
                lines_covered += 1

            if line.get('branch') == "true":
                branch_stats = line.get('condition-coverage')
                b_covered, \
                    b_total = self._parse_branch_stats(branch_stats)
                branches_covered += b_covered
                total_branches += b_total

        self._update_target_stats(**{
            'target': target,
            'lines_covered': lines_covered,
            'total_lines': total_lines,
            'branches_covered': branches_covered,
            'total_branches': total_branches

        })

    @staticmethod
    def _parse_branch_stats(s):
        """Parse branch string from XML: condition-coverage="50% (1/2)" """
        covered = 0
        total = 0
        branch_stats = s[s.find("(")+1:s.find(")")]
        covered, total = branch_stats.split('/')
        return int(covered), int(total)



"""Cobertura aggregation runner"""
import argparse
import imp
from cli_tools import argument
from base import CoberturaXMLAggregator, CoberturaJSONAggregator


@argument('--config', type=str, help="Path to aggregator configuration file")
def run(config):
    """runner"""
    settings_mod = imp.load_source('SETTINGS', config)
    settings_dict = settings_mod.SETTINGS

    # Convert to use python argparse built in
    api_agg = CoberturaJSONAggregator()
    xml_agg = CoberturaXMLAggregator()
    for name, settings in settings_dict.iteritems():
        agg_type = settings.get(CoberturaXMLAggregator.TYPE_KEY)
        if agg_type == 'xml':
            aggregator = xml_agg
        elif agg_type == 'jenkins_api':
            aggregator = api_agg

        aggregator.set_settings(name, settings)
        aggregator.generate_report()
        aggregator.print_report()

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', type=str)
    args = parser.parse_args()
    run(args.config)



"""CoberturaAggregator setup"""
from setuptools import setup, find_packages
setup(
    name="cobertura-aggregator",
    version="0.1",
    packages=find_packages(),
    license="Creative Commons Attribution-Noncommercial-Share Alike license",
    author="Juan Gutierrez",
    description="Cobertura folder coverage aggregator",
    url="https://github.com/juannyg/cobertura_aggregator/",
    install_requires=[
        "tabulate>=0.7.2",
        "simplejson>=3.6.0",
        "cli-tools>=0.2.5"
    ],
    entry_points={
        "console_scripts": [
            "cobertura_agg = cobertura_aggregator.run:run.console"
        ]
    }
)
